#include <malloc.h>
#include <stdio.h>
#include <intrin.h>
#include "..\whirlpool\whirlpool.h"
#include "..\sha3\sha3.h"
#include "scrypt.h"

#ifdef TEST_SHA256
#include "..\sha256\sha256.h"

void hmac_sha256(unsigned char *k, int lk, unsigned char *d, int ld, unsigned char *out)
{
	unsigned char ipad[SHA256_DIGESTSIZE];
	unsigned char key[SHA256_DIGESTSIZE];
	unsigned char buf[SHA256_BLOCKSIZE];
	int i;

	if (lk > SHA256_BLOCKSIZE) {
		sha256(key, k, lk, 1, 1);
		k = key;
		lk = SHA256_DIGESTSIZE;
	}

	for (i = 0; i < lk; i++)
		buf[i] = k[i] ^ 0x36;
	for (; i < SHA256_BLOCKSIZE; i++)
		buf[i] = 0x36;

	sha256(ipad, buf, SHA256_BLOCKSIZE, 1, 0);
	sha256(ipad, d, ld, 0, 1);

	for (i = 0; i < lk; i++)
		buf[i] = k[i] ^ 0x5C;
	for (; i < SHA256_BLOCKSIZE; i++)
		buf[i] = 0x5C;

	sha256(out, buf, SHA256_BLOCKSIZE, 1, 0);
	sha256(out, ipad, SHA256_DIGESTSIZE, 0, 1);

	__stosb(ipad, 0, sizeof ipad);
	__stosb(key, 0, sizeof key);
	__stosb(buf, 0, sizeof buf);
}

static void derive_key_sha256(unsigned char *pwd, int pwd_len, unsigned char *salt, int salt_len, unsigned char *dk, int dklen)
{
	int b, l = dklen / SHA256_DIGESTSIZE;

	for (b = 1; b <= l; b++) {
		// scrypt needs only the 1st iteration of PBKDF2
		*(unsigned long*)(salt + salt_len) = _byteswap_ulong(b);
		hmac_sha256(pwd, pwd_len, salt, salt_len + 4, dk);
		dk += SHA256_DIGESTSIZE;
	}
}
#endif


void hmac_whirlpool(unsigned char *k, int lk, unsigned char *d, int ld, unsigned char *out)
{
	unsigned char ipad[WHIRLPOOL_DIGESTSIZE];
	unsigned char key[WHIRLPOOL_DIGESTSIZE];
	unsigned char buf[WHIRLPOOL_BLOCKSIZE];
	int i;

	if (lk > WHIRLPOOL_BLOCKSIZE) {
		whirlpool(k, lk, key, 1, 1);
		k = key;
		lk = WHIRLPOOL_DIGESTSIZE;
	}

	for (i = 0; i < lk; i++)
		buf[i] = k[i] ^ 0x36;
	for (; i < WHIRLPOOL_BLOCKSIZE; i++)
		buf[i] = 0x36;

	whirlpool(buf, WHIRLPOOL_BLOCKSIZE, ipad, 1, 0);
	whirlpool(d, ld, ipad, 0, 1);

	for (i = 0; i < lk; i++)
		buf[i] = k[i] ^ 0x5C;
	for (; i < WHIRLPOOL_BLOCKSIZE; i++)
		buf[i] = 0x5C;

	whirlpool(buf, WHIRLPOOL_BLOCKSIZE, out, 1, 0);
	whirlpool(ipad, WHIRLPOOL_DIGESTSIZE, out, 0, 1);

	__stosb(ipad, 0, sizeof ipad);
	__stosb(key, 0, sizeof key);
	__stosb(buf, 0, sizeof buf);
}

static void derive_key_whirlpool(unsigned char *pwd, int pwd_len, unsigned char *salt, int salt_len, unsigned char *dk, int dklen)
{
	int b, l = dklen / WHIRLPOOL_DIGESTSIZE;

	for (b = 1; b <= l; b++) {
		// scrypt needs only the 1st iteration of PBKDF2
		*(unsigned long*)(salt + salt_len) = _byteswap_ulong(b);
		hmac_whirlpool(pwd, pwd_len, salt, salt_len + 4, dk);
		dk += WHIRLPOOL_DIGESTSIZE;
	}
}


void hmac_sha3(unsigned char *k, int lk, unsigned char *d, int ld, unsigned char *out)
{
	unsigned char ipad[SHA3_DIGESTSIZE];
	unsigned char key[SHA3_DIGESTSIZE];
	unsigned char buf[SHA3_BLOCKSIZE];
	int i;

	if (lk > SHA3_BLOCKSIZE) {
		sha3(key, k, lk, 1, 1);
		k = key;
		lk = SHA3_DIGESTSIZE;
	}

	for (i = 0; i < lk; i++)
		buf[i] = k[i] ^ 0x36;
	for (; i < SHA3_BLOCKSIZE; i++)
		buf[i] = 0x36;

	sha3(ipad, buf, SHA3_BLOCKSIZE, 1, 0);
	sha3(ipad, d, ld, 0, 1);

	for (i = 0; i < lk; i++)
		buf[i] = k[i] ^ 0x5C;
	for (; i < SHA3_BLOCKSIZE; i++)
		buf[i] = 0x5C;

	sha3(out, buf, SHA3_BLOCKSIZE, 1, 0);
	sha3(out, ipad, SHA3_DIGESTSIZE, 0, 1);

	__stosb(ipad, 0, sizeof ipad);
	__stosb(key, 0, sizeof key);
	__stosb(buf, 0, sizeof buf);
}

static void derive_key_sha3(unsigned char *pwd, int pwd_len, unsigned char *salt, int salt_len, unsigned char *dk, int dklen)
{
	int b, l = dklen / SHA3_DIGESTSIZE;

	for (b = 1; b <= l; b++) {
		// scrypt needs only the 1st iteration of PBKDF2
		*(unsigned long*)(salt + salt_len) = _byteswap_ulong(b);
		hmac_sha3(pwd, pwd_len, salt, salt_len + 4, dk);
		dk += SHA3_DIGESTSIZE;
	}
}


static void blkcpy128(__m128i *dest, __m128i *src, unsigned int len)
{
	unsigned int i;

	for (i = 0; i < len; i++)
		dest[i] = src[i];
}

static void blkxor128(__m128i *dest, __m128i *src, unsigned int len)
{
	unsigned int i;

	for (i = 0; i < len; i++)
		dest[i] ^= src[i];
}

static void salsa20_8(__m128i *B)
{
	__m128i X0, X1, X2, X3, T;
	unsigned int i;

	X0 = B[0];
	X1 = B[1];
	X2 = B[2];
	X3 = B[3];

	for (i = 0; i < 4; i++) {
		T = _mm_add_epi32(X0, X3);
		X1 ^= _mm_slli_epi32(T, 7) ^ _mm_srli_epi32(T, 25);
		T = _mm_add_epi32(X1, X0);
		X2 ^= _mm_slli_epi32(T, 9) ^ _mm_srli_epi32(T, 23);
		T = _mm_add_epi32(X2, X1);
		X3 ^= _mm_slli_epi32(T, 13) ^ _mm_srli_epi32(T, 19);
		T = _mm_add_epi32(X3, X2);
		X0 ^= _mm_slli_epi32(T, 18) ^ _mm_srli_epi32(T, 14);

		X1 = _mm_shuffle_epi32(X1, 0x93);
		X2 = _mm_shuffle_epi32(X2, 0x4E);
		X3 = _mm_shuffle_epi32(X3, 0x39);

		T = _mm_add_epi32(X0, X1);
		X3 ^= _mm_slli_epi32(T, 7) ^ _mm_srli_epi32(T, 25);
		T = _mm_add_epi32(X3, X0);
		X2 ^= _mm_slli_epi32(T, 9) ^ _mm_srli_epi32(T, 23);
		T = _mm_add_epi32(X2, X3);
		X1 ^= _mm_slli_epi32(T, 13) ^ _mm_srli_epi32(T, 19);
		T = _mm_add_epi32(X1, X2);
		X0 ^= _mm_slli_epi32(T, 18) ^ _mm_srli_epi32(T, 14);

		X1 = _mm_shuffle_epi32(X1, 0x39);
		X2 = _mm_shuffle_epi32(X2, 0x4E);
		X3 = _mm_shuffle_epi32(X3, 0x93);
	}

	B[0] = _mm_add_epi32(B[0], X0);
	B[1] = _mm_add_epi32(B[1], X1);
	B[2] = _mm_add_epi32(B[2], X2);
	B[3] = _mm_add_epi32(B[3], X3);
}

static void blockmix_salsa8(__m128i *Bin, __m128i *Bout, __m128i *X, unsigned int r)
{
	unsigned int i;

	blkcpy128(X, &Bin[8 * r - 4], 4);

	for (i = 0; i < r; i++) {
		blkxor128(X, &Bin[i * 8], 4);
		salsa20_8(X);
		blkcpy128(&Bout[i * 4], X, 4);
		blkxor128(X, &Bin[i * 8 + 4], 4);
		salsa20_8(X);
		blkcpy128(&Bout[(r + i) * 4], X, 4);
	}
}

static unsigned long long integerify(void *B, unsigned int r)
{
	unsigned int *X = B + 128 * r - 64;

	return (((unsigned long long)X[13] << 32) + X[0]);
}

static void smix(unsigned char *B, unsigned int r, unsigned long long N, void *V, void *XY)
{
	__m128i *Y = XY + 128 * r;
	__m128i *Z = XY + 256 * r;
	unsigned long long i, j;
	unsigned int k;

	for (k = 0; k < 32 * r; k += 16)
		for (i = 0; i < 16; i++)
			((unsigned int*)XY)[k + i] = ((unsigned int*)B)[k + (i * 5 & 15)];

	for (i = 0; i < N; i += 2) {
		blkcpy128(V + i * 128 * r, XY, 8 * r);
		blockmix_salsa8(XY, Y, Z, r);
		blkcpy128(V + (i + 1) * 128 * r, Y, 8 * r);
		blockmix_salsa8(Y, XY, Z, r);
	}

	for (i = 0; i < N; i += 2) {
		j = integerify(XY, r) & (N - 1);
		blkxor128(XY, V + j * 128 * r, 8 * r);
		blockmix_salsa8(XY, Y, Z, r);
		j = integerify(Y, r) & (N - 1);
		blkxor128(Y, V + j * 128 * r, 8 * r);
		blockmix_salsa8(Y, XY, Z, r);
	}

	for (k = 0; k < 32 * r; k += 16)
		for (i = 0; i < 16; i++)
			((unsigned int*)B)[k + (i * 5 & 15)] = ((unsigned int*)XY)[k + i];
}

int scrypt(int prf, void *passwd, unsigned int passwdlen, void *salt, unsigned int saltlen, unsigned long N, unsigned int r, unsigned int p, void *buf, unsigned int buflen)
// salt must be 4 byte free at its end, saltlen doesn't include these bytes
{
	unsigned char *B;
	unsigned int *V, *XY;
	unsigned int i, size_B, size_V, size_XY;

	size_B = 128 * r * p;
	size_V = 128 * r * N;
	size_XY = 256 * r + 64;
	if (!(B = _aligned_malloc(size_B + 16 + size_V + size_XY, 16))) {
		fputs("Cannot allocate memory for scrypt.", stderr);
		return 1;
	}
	V = (unsigned int*)(B + size_B + 16);
	XY = (unsigned int*)((char*)V + size_V);

#ifdef TEST_SHA256
	if (prf == PRF_SHA256)
		derive_key_sha256(passwd, passwdlen, salt, saltlen, B, size_B);
	else
#endif
	if (prf == PRF_WHIRLPOOL)
		derive_key_whirlpool(passwd, passwdlen, salt, saltlen, B, size_B);
	else
		derive_key_sha3(passwd, passwdlen, salt, saltlen, B, size_B);

	for (i = 0; i < p; i++)
		smix(&B[i * 128 * r], r, N, V, XY);

#ifdef TEST_SHA256
	if (prf == PRF_SHA256)
		derive_key_sha256(passwd, passwdlen, B, size_B, buf, buflen);
	else
#endif
	if (prf == PRF_WHIRLPOOL)
		derive_key_whirlpool(passwd, passwdlen, B, size_B, buf, buflen);
	else
		derive_key_sha3(passwd, passwdlen, B, size_B, buf, buflen);

	__stosb(B, 0, size_B + 16 + size_V + size_XY);
	_aligned_free(B);

	return 0;
}
